#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/kd.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include "names.h"

#define EVDEV          "/dev/input/event"
#define LIRCDEV        "/dev/lircd"
#define PROGNAME       "activylircd"
#define VERSION        "0.3"

#define PLAY_PAUSE_KEY 0x8f
#define MAX_VALUE      0xff

typedef struct client {
   int fd;
   struct client *next;
} client_t;

static client_t *clients = NULL;

typedef struct keycode_lircmap {
   int value;
   int keycode;
   char * lircString;
} key2lirc;

static key2lirc activyKeys[] = {
   { 0x82, KEY_VOLUMEUP, "vol+"},
   { 0x83, KEY_F2, "green"},   
   { 0x84, KEY_UNKNOWN, "tv"},
   { 0x85, KEY_EXIT, "exit"},
   { 0x86, KEY_RIGHT, "right"},
   { 0x87, KEY_EJECTCD, "eject"},
   { 0x88, KEY_FASTFORWARD, "ffwd"},
   { 0x8c, KEY_LEFT, "left"},
   { 0x8e, KEY_LEFT, "kleft"},
   { 0x8f, KEY_PLAY, "play"},
   { 0x90, KEY_POWER, "power"},
   { 0x91, KEY_F4, "blue"},
   { 0x93, KEY_LEFT, "left"},
   { 0x94, KEY_DOWN, "down"},
   { 0x95, KEY_NEXTSONG, "jump+"},
   { 0x96, KEY_SCROLLUP, "scr+"},
   { 0x99, KEY_RIGHT, "right"},
   { 0x9e, KEY_KPMINUS, "ch-"},
   { 0x9f, KEY_KPPLUS, "ch+"},
   { 0xa1, KEY_UP, "up"},
   { 0xa2, KEY_PREVIOUSSONG, "jump-"},
   { 0xa3, KEY_RECORD, "rec"},
   { 0xa4, KEY_POWER, "power"},
   { 0xa8, KEY_KPENTER, "ok"},
   { 0xab, KEY_ESC, "clear"},
   { 0xac, KEY_SCROLLDOWN, "scr-"},
   { 0xae, KEY_MENU, "menu"},
   { 0xaf, KEY_HOME, "home"},
   { 0xb0, KEY_MUTE, "mute"},
   { 0xb1, KEY_HELP, "info"},
   { 0xb2, KEY_DOWN, "down"},
   { 0xb9, KEY_F3, "yellow"},
   { 0xbb, KEY_5, "5"},
   { 0xbc, KEY_6, "6"},
   { 0xbd, KEY_4, "4"},
   { 0xbf, KEY_3, "3"},
   { 0xc1, KEY_2, "2"},
   { 0xc2, KEY_0, "0"},
   { 0xc3, KEY_1, "1"},
   { 0xc4, KEY_9, "9"},
   { 0xd5, KEY_KP8, "dup"},
   { 0xd6, KEY_KP2, "ddown"},
   { 0xd8, KEY_7, "7"},
   { 0xd9, KEY_STOP, "stop"},
   { 0xda, KEY_VOLUMEDOWN, "vol-"},
   { 0xe4, KEY_8, "8"},
   { 0xe7, KEY_F1, "red"},
   { 0xe8, KEY_CONFIG, "option"},
   { 0xe9, KEY_KPENTER, "ok"},
   { 0xea, KEY_REWIND, "frew"},
   { 0xed, KEY_UP, "up"},
   { 0xf5, KEY_CANCEL, "cancel"},
   { 0xf6, KEY_CONFIG, "option"},
   { 0xf7, KEY_KP4, "dleft"},
   { 0xf8, KEY_KP6, "dright"},
   { 0xf9, KEY_UNKNOWN, "tv"},
   { 0xfd, KEY_RIGHT, "kright"},
   { 0xfe, KEY_ENTER, "enter"},
   { 0, 0, NULL}
};

static key2lirc keyTable[ MAX_VALUE + 1 ];

/* prototypes */
static void no_linger(int sock);
static void setup_key_table(void);
static int setup_lirc_socket(void);
static int setup_event_dev(void);
static void process_key(int efd);
static void process_new_client(int sockfd);
static int is_a_console(int fd);
static int open_a_console(const char *fnam);
static int get_console(void);


static void no_linger(int sock)
{
   static struct linger  linger = {0, 0};
   int lsize  = sizeof(struct linger);
   setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, lsize);
}

static void setup_key_table()
{
   int i, conFd;

   for( i = 1; i <= MAX_VALUE; i ++ ) {
      keyTable[i].value = keyTable[i].keycode = 0;
      keyTable[i].lircString = NULL;
   }
   keyTable[0].lircString = "pause";

   for( i = 0; activyKeys[i].value > 0 && activyKeys[i].value <= MAX_VALUE; i++ )
      keyTable[ activyKeys[i].value ] = activyKeys[i];

   /* Setting keycodes */
   conFd = get_console();
   if( conFd >= 0 ) {
      struct kbkeycode a;

      for( i = 0; activyKeys[i].value > 0 && activyKeys[i].value <= MAX_VALUE; i++ ) {
         a.keycode = 120;
         a.scancode = activyKeys[i].value;
         if( ioctl(conFd,KDSETKEYCODE,&a) )
            fprintf(stderr,"failed to set scancode %x to keycode %d\n",
                    a.scancode, a.keycode);
      }
      close(conFd);
   }
}

static int setup_lirc_socket()
{
   int sockfd;
   struct sockaddr_un serv_addr;

   /* create socket */
   sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
   if( sockfd == -1 ) {
      fprintf(stderr, "%s: could not create socket\n", PROGNAME);
      perror(PROGNAME);
      return -1;
   }
   unlink(LIRCDEV);

   serv_addr.sun_family = AF_UNIX;
   strcpy(serv_addr.sun_path, LIRCDEV);
   if( bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) == -1 ) {
      fprintf(stderr, "%s: could not assign address to socket\n",PROGNAME);
      return -1;
   }

   chmod(LIRCDEV, 0666);
   listen(sockfd, 3);
   no_linger(sockfd);

   return sockfd;
}

static int setup_event_dev()
{
   int rfd = -1;
   int i;
   char name[256];
   char info[128];

   for( i = 0; i < 16; i++ ) {
      sprintf( name, "%s%d", EVDEV, i );
      if( (rfd = open (name, O_RDONLY)) >= 0 ) {
         ioctl (rfd, EVIOCGNAME (sizeof (info)), info);
         if( strstr( info, "keyboard" ) ) {
            printf("Using: %s (%s)\n", name, info);
            break;
         }
         printf("Not using: %s (%s)\n", name, info);
         close( rfd );
         rfd = -1;
      }
      else
         fprintf(stderr,"Can't open input event device %s\n", name);
   }
   return rfd;
}


static void process_key(int efd)
{
   static int lastKey = 0;
   struct input_event event[64];
   char message[1000];
   int len, idx;
   int broken = 0;
   client_t *client, *prev, *next;

   len = read(efd, event, sizeof(struct input_event) * 64);

   if( len < (int) sizeof(struct input_event) ) {
      fprintf(stderr, "Error processing input event %s\n", strerror(errno));
      return;
   }

//    for (i = 0; i < (int) (len / sizeof(struct input_event)); i++)
//    {
//       printf("Event: time %ld.%06ld, type %d, code %d, value %d\n",
//              event[i].time.tv_sec, event[i].time.tv_usec, event[i].type,
//              event[i].code, event[i].value);
//
//    }
   if( event[1].value < 1 || event[0].value > MAX_VALUE )
      return;
//
   printf ("%02x 0x%04x %04x\n", event[0].type, event[0].value, event[0].code);
   idx = event[0].value;
   if( keyTable[idx].lircString != NULL ) {
      if( lastKey == PLAY_PAUSE_KEY && keyTable[idx].value == PLAY_PAUSE_KEY )
         lastKey = 0;
      else
         lastKey = keyTable[idx].value;

      len = snprintf(message, sizeof message, "%016x %02x %s ACTIVY\n", event[0].value, 0, keyTable[lastKey].lircString);
   }
   else {
      if( KEY_NAME[event[1].code] )
         len = snprintf(message, sizeof message, "%016x %02x %s ACTIVY\n", event[1].code, 0, KEY_NAME[event[1].code]);
      else
         len = snprintf(message, sizeof message, "%016x %02x KEY_CODE_%d ACTIVY\n", event[1].code, 0, event[1].code);
   }

   for( client = clients; client; client = client->next ) {
      if( write(client->fd, message, len) != len ) {
         close(client->fd);
         client->fd = -1;
         broken = 1;
      }
   }

   if( broken ) {
      for( prev = NULL, client = clients; client; client = next ) {
         next = client->next;
         if( client->fd < 0 ) {
            if( prev )
               prev->next = client->next;
            else
               clients = client->next;
            free(client);
         }
         else {
            prev = client;
         }
      }
   }
   return;
}

static void process_new_client(int sockfd) 
{
   client_t *newclient = calloc(sizeof *newclient, 1);

   newclient->fd = accept(sockfd, NULL, NULL);

   if( newclient->fd < 0 ) {
      free(newclient);
      if( errno == ECONNABORTED || errno == EINTR )
         return;
      fprintf(stderr, "Error during accept(): %s\n", strerror(errno));
      return;
   }

   int flags = fcntl(newclient->fd, F_GETFL);
   fcntl(newclient->fd, F_SETFL, flags | O_NONBLOCK);
   no_linger(newclient->fd);

   newclient->next = clients;
   clients = newclient;
}


int main()
{
   fd_set permset, fdset;
   int max_fd, event_fd, lirc_fd;

   printf( "\n%s Version %s\n", PROGNAME, VERSION );
   event_fd = setup_event_dev();
   if( event_fd < 0 ) {
      fprintf(stderr,"No input device found\n" );
      return 1;
   }

   setup_key_table();

   //GetKey(rfd, message);
   lirc_fd = setup_lirc_socket();
   if( lirc_fd < 0 ) {
      fprintf(stderr,"Can't init lirc socket.\n");
      return -1;
   }

   signal(SIGPIPE, SIG_IGN);

   FD_ZERO(&permset);
   FD_SET(event_fd, &permset);
   FD_SET(lirc_fd, &permset);
   if( event_fd > lirc_fd )
      max_fd = event_fd;
   else
      max_fd = lirc_fd;

   max_fd++;

   while( 1 ) {
      fdset = permset;

      if( select(max_fd, &fdset, NULL, NULL, NULL) < 0 ) {
         if( errno == EINTR )
            continue;
         fprintf(stderr, "Error during select(): %s\n", strerror(errno));
         return -1;
      }

      if( FD_ISSET(event_fd, &fdset) )
         process_key(event_fd);

      if( FD_ISSET(lirc_fd, &fdset) )
         process_new_client(lirc_fd);
   }

   unlink(LIRCDEV);
   return 0;
}


static int is_a_console(int fd) 
{
   char arg;

   arg = 0;
   return(ioctl(fd, KDGKBTYPE, &arg) == 0
          && ((arg == KB_101) || (arg == KB_84)));
}

static int open_a_console(const char *fnam) 
{
   int fd;

   /*
    * For ioctl purposes we only need some fd and permissions
    * do not matter. But setfont:activatemap() does a write.
    */
   fd = open(fnam, O_RDWR);
   if( fd < 0 && errno == EACCES )
      fd = open(fnam, O_WRONLY);
   if( fd < 0 && errno == EACCES )
      fd = open(fnam, O_RDONLY);
   if( fd < 0 )
      return -1;
   if( !is_a_console(fd) ) {
      close(fd);
      return -1;
   }
   return fd;
}

static int get_console( void )
{
   int fd;

   fd = open_a_console("/dev/tty");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/tty0");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/vc/0");
   if( fd >= 0 )
      return fd;

   fd = open_a_console("/dev/console");
   if( fd >= 0 )
      return fd;

   for( fd = 0; fd < 3; fd++ )
      if( is_a_console(fd) )
         return fd;

   fprintf(stderr, "Couldnt get a file descriptor referring to the console\n");
   return -1;
}
